/***
  This file is part of PulseAudio.

  Copyright 2004-2006 Lennart Poettering
  Copyright 2006 Pierre Ossman <ossman@cendio.se> for Cendio AB

  PulseAudio is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation; either version 2.1 of the License,
  or (at your option) any later version.

  PulseAudio is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with PulseAudio; if not, see <http://www.gnu.org/licenses/>.
***/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>

#include <pulse/rtclock.h>
#include <pulse/timeval.h>
#include <pulse/xmalloc.h>

#include <pulsecore/module.h>
#include <pulsecore/core-struct.h>
#include <pulsecore/core-rtclock.h>
#include <pulsecore/core-util.h>
#include <pulsecore/core-scache.h>
#include <pulsecore/core-subscribe.h>
#include <pulsecore/random.h>
#include <pulsecore/log.h>
#include <pulsecore/macro.h>

#include "core.h"

PA_DEFINE_PUBLIC_CLASS(pa_core, pa_msgobject);

static int core_process_msg(pa_msgobject *o, int code, void *userdata, int64_t offset, pa_memchunk *chunk) {
    pa_core *c = PA_CORE(o);

    pa_core_assert_ref(c);

    switch (code) {

        case PA_CORE_MESSAGE_UNLOAD_MODULE:
            pa_module_unload(userdata, true);
            return 0;

        default:
            return -1;
    }
}

static void core_free(pa_object *o);

pa_core* pa_core_new(pa_mainloop_api *m, bool shared, bool enable_memfd, size_t shm_size) {
    pa_core* c;
    pa_mempool *pool;
    pa_mem_type_t type;
    int j;

    pa_assert(m);

    if (shared) {
        type = (enable_memfd) ? PA_MEM_TYPE_SHARED_MEMFD : PA_MEM_TYPE_SHARED_POSIX;
        if (!(pool = pa_mempool_new(type, shm_size, false))) {
            pa_log_warn("Failed to allocate %s memory pool. Falling back to a normal memory pool.",
                        pa_mem_type_to_string(type));
            shared = false;
        }
    }

    if (!shared) {
        if (!(pool = pa_mempool_new(PA_MEM_TYPE_PRIVATE, shm_size, false))) {
            pa_log("pa_mempool_new() failed.");
            return NULL;
        }
    }

    c = pa_msgobject_new(pa_core);
    c->parent.parent.free = core_free;
    c->parent.process_msg = core_process_msg;

    c->state = PA_CORE_STARTUP;
    c->mainloop = m;

    c->clients = pa_idxset_new(NULL, NULL);
    c->cards = pa_idxset_new(NULL, NULL);
    c->sinks = pa_idxset_new(NULL, NULL);
    c->sources = pa_idxset_new(NULL, NULL);
    c->sink_inputs = pa_idxset_new(NULL, NULL);
    c->source_outputs = pa_idxset_new(NULL, NULL);
    c->modules = pa_idxset_new(NULL, NULL);
    c->scache = pa_idxset_new(NULL, NULL);

    c->namereg = pa_hashmap_new(pa_idxset_string_hash_func, pa_idxset_string_compare_func);
    c->shared = pa_hashmap_new(pa_idxset_string_hash_func, pa_idxset_string_compare_func);

    c->default_source = NULL;
    c->default_sink = NULL;

    c->default_sample_spec.format = PA_SAMPLE_S16NE;
    c->default_sample_spec.rate = 44100;
    c->default_sample_spec.channels = 2;
    pa_channel_map_init_extend(&c->default_channel_map, c->default_sample_spec.channels, PA_CHANNEL_MAP_DEFAULT);
    c->default_n_fragments = 4;
    c->default_fragment_size_msec = 25;

    c->deferred_volume_safety_margin_usec = 8000;
    c->deferred_volume_extra_delay_usec = 0;

    c->module_defer_unload_event = NULL;
    c->modules_pending_unload = pa_hashmap_new(NULL, NULL);

    c->subscription_defer_event = NULL;
    PA_LLIST_HEAD_INIT(pa_subscription, c->subscriptions);
    PA_LLIST_HEAD_INIT(pa_subscription_event, c->subscription_event_queue);
    c->subscription_event_last = NULL;

    c->mempool = pool;
    c->shm_size = shm_size;
    pa_silence_cache_init(&c->silence_cache);

    c->exit_event = NULL;
    c->scache_auto_unload_event = NULL;

    c->exit_idle_time = -1;
    c->scache_idle_time = 20;

    c->flat_volumes = false;
    c->disallow_module_loading = false;
    c->disallow_exit = false;
    c->running_as_daemon = false;
    c->realtime_scheduling = false;
    c->realtime_priority = 5;
    c->disable_remixing = false;
    c->disable_lfe_remixing = true;
    c->lfe_crossover_freq = 0;
    c->deferred_volume = true;
    c->resample_method = PA_RESAMPLER_SPEEX_FLOAT_BASE + 1;

    for (j = 0; j < PA_CORE_HOOK_MAX; j++)
        pa_hook_init(&c->hooks[j], c);

    for (j = 0; j < PA_ACCESS_HOOK_MAX; j++)
        pa_hook_init(&c->access[j], c);

    c->current_client = NULL;

    pa_random(&c->cookie, sizeof(c->cookie));

#ifdef SIGPIPE
    pa_check_signal_is_blocked(SIGPIPE);
#endif

    pa_core_check_idle(c);

    c->state = PA_CORE_RUNNING;

    return c;
}

static void core_free(pa_object *o) {
    pa_core *c = PA_CORE(o);
    int j;
    pa_assert(c);

    c->state = PA_CORE_SHUTDOWN;

    /* Note: All modules and samples in the cache should be unloaded before
     * we get here */

    pa_assert(pa_idxset_isempty(c->scache));
    pa_idxset_free(c->scache, NULL);

    pa_assert(pa_idxset_isempty(c->modules));
    pa_idxset_free(c->modules, NULL);

    pa_assert(pa_idxset_isempty(c->clients));
    pa_idxset_free(c->clients, NULL);

    pa_assert(pa_idxset_isempty(c->cards));
    pa_idxset_free(c->cards, NULL);

    pa_assert(pa_idxset_isempty(c->sinks));
    pa_idxset_free(c->sinks, NULL);

    pa_assert(pa_idxset_isempty(c->sources));
    pa_idxset_free(c->sources, NULL);

    pa_assert(pa_idxset_isempty(c->source_outputs));
    pa_idxset_free(c->source_outputs, NULL);

    pa_assert(pa_idxset_isempty(c->sink_inputs));
    pa_idxset_free(c->sink_inputs, NULL);

    pa_assert(pa_hashmap_isempty(c->namereg));
    pa_hashmap_free(c->namereg);

    pa_assert(pa_hashmap_isempty(c->shared));
    pa_hashmap_free(c->shared);

    pa_assert(pa_hashmap_isempty(c->modules_pending_unload));
    pa_hashmap_free(c->modules_pending_unload);

    pa_subscription_free_all(c);

    if (c->exit_event)
        c->mainloop->time_free(c->exit_event);

    pa_assert(!c->default_source);
    pa_assert(!c->default_sink);

    pa_silence_cache_done(&c->silence_cache);
    pa_mempool_unref(c->mempool);

    for (j = 0; j < PA_CORE_HOOK_MAX; j++)
        pa_hook_done(&c->hooks[j]);
    for (j = 0; j < PA_ACCESS_HOOK_MAX; j++)
        pa_hook_done(&c->access[j]);

    pa_xfree(c);
}

static void exit_callback(pa_mainloop_api *m, pa_time_event *e, const struct timeval *t, void *userdata) {
    pa_core *c = userdata;
    pa_assert(c->exit_event == e);

    pa_log_info("We are idle, quitting...");
    pa_core_exit(c, true, 0);
}

void pa_core_check_idle(pa_core *c) {
    pa_assert(c);

    if (!c->exit_event &&
        c->exit_idle_time >= 0 &&
        pa_idxset_size(c->clients) == 0) {

        c->exit_event = pa_core_rttime_new(c, pa_rtclock_now() + c->exit_idle_time * PA_USEC_PER_SEC, exit_callback, c);

    } else if (c->exit_event && pa_idxset_size(c->clients) > 0) {
        c->mainloop->time_free(c->exit_event);
        c->exit_event = NULL;
    }
}

int pa_core_exit(pa_core *c, bool force, int retval) {
    pa_assert(c);

    if (c->disallow_exit && !force)
        return -1;
    if (pa_core_check_access_sync(c, PA_ACCESS_HOOK_EXIT_DAEMON, PA_INVALID_INDEX, 0, NULL))
        return -1;

    c->mainloop->quit(c->mainloop, retval);
    return 0;
}

pa_hook_result_t pa_core_hook_fire(pa_core *c, pa_core_hook_t hook, void *data) {
    pa_assert(c);
    return pa_hook_fire(&c->hooks[hook], data);
}

void pa_core_maybe_vacuum(pa_core *c) {
    pa_assert(c);

    if (pa_idxset_isempty(c->sink_inputs) && pa_idxset_isempty(c->source_outputs)) {
        pa_log_debug("Hmm, no streams around, trying to vacuum.");
    } else {
        pa_sink *si;
        pa_source *so;
        uint32_t idx;

        idx = 0;
        PA_IDXSET_FOREACH(si, c->sinks, idx)
            if (pa_sink_get_state(si) != PA_SINK_SUSPENDED)
                return;

        idx = 0;
        PA_IDXSET_FOREACH(so, c->sources, idx)
            if (pa_source_get_state(so) != PA_SOURCE_SUSPENDED)
                return;

        pa_log_info("All sinks and sources are suspended, vacuuming memory");
    }

    pa_mempool_vacuum(c->mempool);
}

pa_time_event* pa_core_rttime_new(pa_core *c, pa_usec_t usec, pa_time_event_cb_t cb, void *userdata) {
    struct timeval tv;

    pa_assert(c);
    pa_assert(c->mainloop);

    return c->mainloop->time_new(c->mainloop, pa_timeval_rtstore(&tv, usec, true), cb, userdata);
}

void pa_core_rttime_restart(pa_core *c, pa_time_event *e, pa_usec_t usec) {
    struct timeval tv;

    pa_assert(c);
    pa_assert(c->mainloop);

    c->mainloop->time_restart(e, pa_timeval_rtstore(&tv, usec, true));
}

pa_mainloop_api* pa_core_get_mainloop(pa_core *c) {
    pa_assert(c);
    pa_assert(c->mainloop);

    return c->mainloop;
}

/* FIXME: Should we expose this at all? */
pa_mempool* pa_core_get_mempool(pa_core *c) {
    pa_assert(c);
    pa_assert(c->mempool);

    return c->mempool;
}

pa_mempool* pa_core_new_mempool(pa_core *c, pa_mem_type_t shm_type, bool per_client) {
    return pa_mempool_new(shm_type, c->shm_size, per_client);
}

pa_client* pa_core_get_current_client(pa_core *c) {
    pa_assert_ctl_context();
    return c->current_client;
}

void pa_core_set_current_client(pa_core *c, pa_client *client) {
    pa_assert_ctl_context();
    c->current_client = client;
}

bool pa_core_check_access_sync(pa_core *c, pa_access_hook_t hook, uint32_t idx, pa_subscription_event_type_t event, const char *name) {
    pa_access_data data;

    pa_assert(c);
    pa_assert_ctl_context();

    if (c->current_client == NULL)
      return true;

    data.hook = hook;
    data.client_index = c->current_client->index;
    data.object_index = idx;
    data.event = event;
    data.name = name;
    data.complete_cb = NULL;

    return pa_hook_fire(&c->access[data.hook], &data) == PA_HOOK_OK;
}

pa_hook_result_t pa_core_check_access(pa_core *c, pa_access_data *data) {
    pa_assert(c);
    pa_assert(data);
    pa_assert_ctl_context();

    if (c->current_client == NULL)
      return PA_HOOK_OK;

    data->client_index = c->current_client->index;

    return pa_hook_fire(&c->access[data->hook], data);
}

/* FIXME: Should these be taking a ref during the copy? */

pa_idxset* pa_core_get_modules(pa_core *c) {
    pa_idxset *s;
    uint32_t idx;
    pa_module *e;

    s = pa_idxset_new(NULL, NULL);
    PA_IDXSET_FOREACH(e, c->modules, idx) {
        if (pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_MODULE, e->index, 0, NULL))
            pa_idxset_put(s, e, NULL);
    }
    return s;
}

pa_idxset* pa_core_get_clients(pa_core *c) {
    pa_idxset *s;
    uint32_t idx;
    pa_client *e;

    s = pa_idxset_new(NULL, NULL);
    PA_IDXSET_FOREACH(e, c->clients, idx) {
        if (pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_CLIENT, e->index, 0, NULL))
            pa_idxset_put(s, e, NULL);
    }
    return s;
}

pa_idxset* pa_core_get_cards(pa_core *c) {
    pa_idxset *s;
    uint32_t idx;
    pa_card *e;

    s = pa_idxset_new(NULL, NULL);
    PA_IDXSET_FOREACH(e, c->cards, idx) {
        if (pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_CARD, e->index, 0, NULL))
            pa_idxset_put(s, e, NULL);
    }
    return s;
}

pa_idxset* pa_core_get_sinks(pa_core *c) {
    pa_idxset *s;
    uint32_t idx;
    pa_sink *e;

    s = pa_idxset_new(NULL, NULL);
    PA_IDXSET_FOREACH(e, c->sinks, idx) {
        if (pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_SINK, e->index, 0, NULL))
            pa_idxset_put(s, e, NULL);
    }
    return s;
}

pa_idxset* pa_core_get_sources(pa_core *c) {
    pa_idxset *s;
    uint32_t idx;
    pa_source *e;

    s = pa_idxset_new(NULL, NULL);
    PA_IDXSET_FOREACH(e, c->sources, idx) {
        if (pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_SOURCE, e->index, 0, NULL))
            pa_idxset_put(s, e, NULL);
    }
    return s;
}

pa_idxset* pa_core_get_sink_inputs(pa_core *c) {
    pa_idxset *s;
    uint32_t idx;
    pa_sink_input *e;

    s = pa_idxset_new(NULL, NULL);
    PA_IDXSET_FOREACH(e, c->sink_inputs, idx) {
        if (pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_SINK_INPUT, e->index, 0, NULL))
            pa_idxset_put(s, e, NULL);
    }
    return s;
}

pa_idxset* pa_core_get_source_outputs(pa_core *c) {
    pa_idxset *s;
    uint32_t idx;
    pa_source_output *e;

    s = pa_idxset_new(NULL, NULL);
    PA_IDXSET_FOREACH(e, c->source_outputs, idx) {
        if (pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_SOURCE_OUTPUT, e->index, 0, NULL))
            pa_idxset_put(s, e, NULL);
    }
    return s;
}

pa_idxset* pa_core_get_scache(pa_core *c) {
    pa_idxset *s;
    uint32_t idx;
    pa_scache_entry *e;

    s = pa_idxset_new(NULL, NULL);
    PA_IDXSET_FOREACH(e, c->scache, idx) {
        if (pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_SAMPLE, e->index, 0, NULL))
            pa_idxset_put(s, e, NULL);
    }
    return s;
}

pa_module* pa_core_get_module(pa_core *c, uint32_t idx) {
    pa_module *e = pa_idxset_get_by_index(c->modules, idx);

    if (e == NULL || !pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_MODULE, e->index, 0, NULL))
      return NULL;

    return e;
}

pa_client* pa_core_get_client(pa_core *c, uint32_t idx) {
    pa_client *e = pa_idxset_get_by_index(c->clients, idx);

    if (e == NULL || !pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_CLIENT, e->index, 0, NULL))
      return NULL;

    return e;
}

pa_card* pa_core_get_card(pa_core *c, uint32_t idx) {
    pa_card *e = pa_idxset_get_by_index(c->cards, idx);

    if (e == NULL || !pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_CARD, e->index, 0, NULL))
      return NULL;

    return e;
}

pa_sink* pa_core_get_sink(pa_core *c, uint32_t idx) {
    pa_sink *e = pa_idxset_get_by_index(c->sinks, idx);

    if (e == NULL || !pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_SINK, e->index, 0, NULL))
      return NULL;

    return e;
}

pa_source* pa_core_get_source(pa_core *c, uint32_t idx) {
    pa_source *e = pa_idxset_get_by_index(c->sources, idx);

    if (e == NULL || !pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_SOURCE, e->index, 0, NULL))
      return NULL;

    return e;
}

pa_sink_input* pa_core_get_sink_input(pa_core *c, uint32_t idx) {
    pa_sink_input *e = pa_idxset_get_by_index(c->sink_inputs, idx);

    if (e == NULL || !pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_SINK_INPUT, e->index, 0, NULL))
      return NULL;

    return e;
}

pa_source_output* pa_core_get_source_output(pa_core *c, uint32_t idx) {
    pa_source_output *e = pa_idxset_get_by_index(c->source_outputs, idx);

    if (e == NULL || !pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_SOURCE_OUTPUT, e->index, 0, NULL))
      return NULL;

    return e;
}

pa_scache_entry* pa_core_get_scache_entry(pa_core *c, uint32_t idx) {
    pa_scache_entry *e = pa_idxset_get_by_index(c->scache, idx);

    if (e == NULL || !pa_core_check_access_sync(c, PA_ACCESS_HOOK_VIEW_SAMPLE, e->index, 0, NULL))
      return NULL;

    return e;
}

const pa_sample_spec* pa_core_get_sample_spec(pa_core *c) {
    return &c->default_sample_spec;
}

const pa_channel_map* pa_core_get_channel_map(pa_core *c) {
    return &c->default_channel_map;
}

uint32_t pa_core_get_cookie(pa_core *c) {
    return c->cookie;
}
